"""Generated wrapper for Oracle Solidity contract."""

# pylint: disable=too-many-arguments

import json
from typing import (  # pylint: disable=unused-import
    Any,
    List,
    Optional,
    Tuple,
    Union,
)
import time
from eth_utils import to_checksum_address
from mypy_extensions import TypedDict  # pylint: disable=unused-import
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import ContractFunction
from web3.datastructures import AttributeDict
from web3.providers.base import BaseProvider
from web3.exceptions import ContractLogicError
from moody.m.bases import ContractMethod, Validator, ContractBase, Signatures
from moody.m.tx_params import TxParams
from moody.libeb import MiliDoS
from moody import Bolors

# Try to import a custom validator class definition; if there isn't one,
# declare one that we can instantiate for the default argument to the
# constructor for Oracle below.
try:
    # both mypy and pylint complain about what we're doing here, but this
    # works just fine, so their messages have been disabled here.
    from . import (  # type: ignore # pylint: disable=import-self
        OracleValidator,
    )
except ImportError:

    class OracleValidator(  # type: ignore
        Validator
    ):
        """No-op input validator."""

try:
    from .middleware import MIDDLEWARE  # type: ignore
except ImportError:
    pass


class ExpiryTimeMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the EXPIRY_TIME method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("EXPIRY_TIME")

    def block_call(self, debug: bool = False) -> int:
        _fn = self._underlying_method()
        returned = _fn.call({
            'from': self._operate
        })
        return int(returned)

    def block_send(self, _gaswei: int, _pricewei: int, _valeth: int = 0, _debugtx: bool = False, _receipList: bool = False) -> int:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method()
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': _gaswei,
                'gasPrice': _pricewei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if _debugtx:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if _receipList is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if _debugtx:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET} - broadcast hash")

            if _receipList is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: expiry_time")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, expiry_time: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, expiry_time. Reason: Unknown")

    def send_transaction(self, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())


class CancelOracleRequestMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the cancelOracleRequest method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("cancelOracleRequest")

    def validate_and_normalize_inputs(self, request_id: Union[bytes, str], payment: int, callback_func: Union[bytes, str], expiration: int) -> any:
        """Validate the inputs to the cancelOracleRequest method."""
        self.validator.assert_valid(
            method_name='cancelOracleRequest',
            parameter_name='_requestId',
            argument_value=request_id,
        )
        self.validator.assert_valid(
            method_name='cancelOracleRequest',
            parameter_name='_payment',
            argument_value=payment,
        )
        # safeguard against fractional inputs
        payment = int(payment)
        self.validator.assert_valid(
            method_name='cancelOracleRequest',
            parameter_name='_callbackFunc',
            argument_value=callback_func,
        )
        self.validator.assert_valid(
            method_name='cancelOracleRequest',
            parameter_name='_expiration',
            argument_value=expiration,
        )
        # safeguard against fractional inputs
        expiration = int(expiration)
        return (request_id, payment, callback_func, expiration)

    def block_send(self, request_id: Union[bytes, str], payment: int, callback_func: Union[bytes, str], expiration: int, _gaswei: int, _pricewei: int, _valeth: int = 0, _debugtx: bool = False, _receipList: bool = False) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method(request_id, payment, callback_func, expiration)
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': _gaswei,
                'gasPrice': _pricewei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if _debugtx:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if _receipList is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if _debugtx:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET} - broadcast hash")

            if _receipList is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: cancel_oracle_request")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, cancel_oracle_request: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, cancel_oracle_request. Reason: Unknown")

    def send_transaction(self, request_id: Union[bytes, str], payment: int, callback_func: Union[bytes, str], expiration: int, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (request_id, payment, callback_func, expiration) = self.validate_and_normalize_inputs(request_id, payment, callback_func, expiration)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(request_id, payment, callback_func, expiration).transact(tx_params.as_dict())

    def build_transaction(self, request_id: Union[bytes, str], payment: int, callback_func: Union[bytes, str], expiration: int, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (request_id, payment, callback_func, expiration) = self.validate_and_normalize_inputs(request_id, payment, callback_func, expiration)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(request_id, payment, callback_func, expiration).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, request_id: Union[bytes, str], payment: int, callback_func: Union[bytes, str], expiration: int, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (request_id, payment, callback_func, expiration) = self.validate_and_normalize_inputs(request_id, payment, callback_func, expiration)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(request_id, payment, callback_func, expiration).estimateGas(tx_params.as_dict())


class FulfillOracleRequestMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the fulfillOracleRequest method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("fulfillOracleRequest")

    def validate_and_normalize_inputs(self, request_id: Union[bytes, str], payment: int, callback_address: str, callback_function_id: Union[bytes, str], expiration: int, data: Union[bytes, str]) -> any:
        """Validate the inputs to the fulfillOracleRequest method."""
        self.validator.assert_valid(
            method_name='fulfillOracleRequest',
            parameter_name='_requestId',
            argument_value=request_id,
        )
        self.validator.assert_valid(
            method_name='fulfillOracleRequest',
            parameter_name='_payment',
            argument_value=payment,
        )
        # safeguard against fractional inputs
        payment = int(payment)
        self.validator.assert_valid(
            method_name='fulfillOracleRequest',
            parameter_name='_callbackAddress',
            argument_value=callback_address,
        )
        callback_address = self.validate_and_checksum_address(callback_address)
        self.validator.assert_valid(
            method_name='fulfillOracleRequest',
            parameter_name='_callbackFunctionId',
            argument_value=callback_function_id,
        )
        self.validator.assert_valid(
            method_name='fulfillOracleRequest',
            parameter_name='_expiration',
            argument_value=expiration,
        )
        # safeguard against fractional inputs
        expiration = int(expiration)
        self.validator.assert_valid(
            method_name='fulfillOracleRequest',
            parameter_name='_data',
            argument_value=data,
        )
        return (request_id, payment, callback_address, callback_function_id, expiration, data)

    def block_send(self, request_id: Union[bytes, str], payment: int, callback_address: str, callback_function_id: Union[bytes, str], expiration: int, data: Union[bytes, str], _gaswei: int, _pricewei: int, _valeth: int = 0, _debugtx: bool = False, _receipList: bool = False) -> bool:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method(request_id, payment, callback_address, callback_function_id, expiration, data)
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': _gaswei,
                'gasPrice': _pricewei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if _debugtx:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if _receipList is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if _debugtx:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET} - broadcast hash")

            if _receipList is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: fulfill_oracle_request")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, fulfill_oracle_request: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, fulfill_oracle_request. Reason: Unknown")

    def send_transaction(self, request_id: Union[bytes, str], payment: int, callback_address: str, callback_function_id: Union[bytes, str], expiration: int, data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (request_id, payment, callback_address, callback_function_id, expiration, data) = self.validate_and_normalize_inputs(request_id, payment, callback_address, callback_function_id, expiration, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(request_id, payment, callback_address, callback_function_id, expiration, data).transact(tx_params.as_dict())

    def build_transaction(self, request_id: Union[bytes, str], payment: int, callback_address: str, callback_function_id: Union[bytes, str], expiration: int, data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (request_id, payment, callback_address, callback_function_id, expiration, data) = self.validate_and_normalize_inputs(request_id, payment, callback_address, callback_function_id, expiration, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(request_id, payment, callback_address, callback_function_id, expiration, data).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, request_id: Union[bytes, str], payment: int, callback_address: str, callback_function_id: Union[bytes, str], expiration: int, data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (request_id, payment, callback_address, callback_function_id, expiration, data) = self.validate_and_normalize_inputs(request_id, payment, callback_address, callback_function_id, expiration, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(request_id, payment, callback_address, callback_function_id, expiration, data).estimateGas(tx_params.as_dict())


class GetAuthorizationStatusMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the getAuthorizationStatus method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("getAuthorizationStatus")

    def validate_and_normalize_inputs(self, node: str) -> any:
        """Validate the inputs to the getAuthorizationStatus method."""
        self.validator.assert_valid(
            method_name='getAuthorizationStatus',
            parameter_name='_node',
            argument_value=node,
        )
        node = self.validate_and_checksum_address(node)
        return (node)

    def block_call(self, node: str, debug: bool = False) -> bool:
        _fn = self._underlying_method(node)
        returned = _fn.call({
            'from': self._operate
        })
        return bool(returned)

    def block_send(self, node: str, _gaswei: int, _pricewei: int, _valeth: int = 0, _debugtx: bool = False, _receipList: bool = False) -> bool:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method(node)
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': _gaswei,
                'gasPrice': _pricewei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if _debugtx:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if _receipList is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if _debugtx:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET} - broadcast hash")

            if _receipList is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: get_authorization_status")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, get_authorization_status: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, get_authorization_status. Reason: Unknown")

    def send_transaction(self, node: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (node) = self.validate_and_normalize_inputs(node)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(node).transact(tx_params.as_dict())

    def build_transaction(self, node: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (node) = self.validate_and_normalize_inputs(node)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(node).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, node: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (node) = self.validate_and_normalize_inputs(node)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(node).estimateGas(tx_params.as_dict())


class GetChainlinkTokenMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the getChainlinkToken method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("getChainlinkToken")

    def block_call(self, debug: bool = False) -> str:
        _fn = self._underlying_method()
        returned = _fn.call({
            'from': self._operate
        })
        return str(returned)

    def block_send(self, _gaswei: int, _pricewei: int, _valeth: int = 0, _debugtx: bool = False, _receipList: bool = False) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method()
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': _gaswei,
                'gasPrice': _pricewei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if _debugtx:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if _receipList is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if _debugtx:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET} - broadcast hash")

            if _receipList is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: get_chainlink_token")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, get_chainlink_token: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, get_chainlink_token. Reason: Unknown")

    def send_transaction(self, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())


class IsOwnerMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the isOwner method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("isOwner")

    def block_call(self, debug: bool = False) -> bool:
        _fn = self._underlying_method()
        returned = _fn.call({
            'from': self._operate
        })
        return bool(returned)

    def block_send(self, _gaswei: int, _pricewei: int, _valeth: int = 0, _debugtx: bool = False, _receipList: bool = False) -> bool:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method()
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': _gaswei,
                'gasPrice': _pricewei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if _debugtx:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if _receipList is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if _debugtx:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET} - broadcast hash")

            if _receipList is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: is_owner")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, is_owner: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, is_owner. Reason: Unknown")

    def send_transaction(self, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())


class OnTokenTransferMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the onTokenTransfer method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("onTokenTransfer")

    def validate_and_normalize_inputs(self, sender: str, amount: int, data: Union[bytes, str]) -> any:
        """Validate the inputs to the onTokenTransfer method."""
        self.validator.assert_valid(
            method_name='onTokenTransfer',
            parameter_name='_sender',
            argument_value=sender,
        )
        sender = self.validate_and_checksum_address(sender)
        self.validator.assert_valid(
            method_name='onTokenTransfer',
            parameter_name='_amount',
            argument_value=amount,
        )
        # safeguard against fractional inputs
        amount = int(amount)
        self.validator.assert_valid(
            method_name='onTokenTransfer',
            parameter_name='_data',
            argument_value=data,
        )
        return (sender, amount, data)

    def block_send(self, sender: str, amount: int, data: Union[bytes, str], _gaswei: int, _pricewei: int, _valeth: int = 0, _debugtx: bool = False, _receipList: bool = False) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method(sender, amount, data)
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': _gaswei,
                'gasPrice': _pricewei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if _debugtx:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if _receipList is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if _debugtx:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET} - broadcast hash")

            if _receipList is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: on_token_transfer")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, on_token_transfer: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, on_token_transfer. Reason: Unknown")

    def send_transaction(self, sender: str, amount: int, data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (sender, amount, data) = self.validate_and_normalize_inputs(sender, amount, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(sender, amount, data).transact(tx_params.as_dict())

    def build_transaction(self, sender: str, amount: int, data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (sender, amount, data) = self.validate_and_normalize_inputs(sender, amount, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(sender, amount, data).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, sender: str, amount: int, data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (sender, amount, data) = self.validate_and_normalize_inputs(sender, amount, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(sender, amount, data).estimateGas(tx_params.as_dict())


class OracleRequestMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the oracleRequest method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("oracleRequest")

    def validate_and_normalize_inputs(self, sender: str, payment: int, spec_id: Union[bytes, str], callback_address: str, callback_function_id: Union[bytes, str], nonce: int, data_version: int, data: Union[bytes, str]) -> any:
        """Validate the inputs to the oracleRequest method."""
        self.validator.assert_valid(
            method_name='oracleRequest',
            parameter_name='_sender',
            argument_value=sender,
        )
        sender = self.validate_and_checksum_address(sender)
        self.validator.assert_valid(
            method_name='oracleRequest',
            parameter_name='_payment',
            argument_value=payment,
        )
        # safeguard against fractional inputs
        payment = int(payment)
        self.validator.assert_valid(
            method_name='oracleRequest',
            parameter_name='_specId',
            argument_value=spec_id,
        )
        self.validator.assert_valid(
            method_name='oracleRequest',
            parameter_name='_callbackAddress',
            argument_value=callback_address,
        )
        callback_address = self.validate_and_checksum_address(callback_address)
        self.validator.assert_valid(
            method_name='oracleRequest',
            parameter_name='_callbackFunctionId',
            argument_value=callback_function_id,
        )
        self.validator.assert_valid(
            method_name='oracleRequest',
            parameter_name='_nonce',
            argument_value=nonce,
        )
        # safeguard against fractional inputs
        nonce = int(nonce)
        self.validator.assert_valid(
            method_name='oracleRequest',
            parameter_name='_dataVersion',
            argument_value=data_version,
        )
        # safeguard against fractional inputs
        data_version = int(data_version)
        self.validator.assert_valid(
            method_name='oracleRequest',
            parameter_name='_data',
            argument_value=data,
        )
        return (sender, payment, spec_id, callback_address, callback_function_id, nonce, data_version, data)

    def block_send(self, sender: str, payment: int, spec_id: Union[bytes, str], callback_address: str, callback_function_id: Union[bytes, str], nonce: int, data_version: int, data: Union[bytes, str], _gaswei: int, _pricewei: int, _valeth: int = 0, _debugtx: bool = False, _receipList: bool = False) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method(sender, payment, spec_id, callback_address, callback_function_id, nonce, data_version, data)
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': _gaswei,
                'gasPrice': _pricewei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if _debugtx:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if _receipList is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if _debugtx:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET} - broadcast hash")

            if _receipList is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: oracle_request")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, oracle_request: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, oracle_request. Reason: Unknown")

    def send_transaction(self, sender: str, payment: int, spec_id: Union[bytes, str], callback_address: str, callback_function_id: Union[bytes, str], nonce: int, data_version: int, data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (sender, payment, spec_id, callback_address, callback_function_id, nonce, data_version, data) = self.validate_and_normalize_inputs(sender, payment, spec_id, callback_address, callback_function_id, nonce, data_version, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(sender, payment, spec_id, callback_address, callback_function_id, nonce, data_version, data).transact(tx_params.as_dict())

    def build_transaction(self, sender: str, payment: int, spec_id: Union[bytes, str], callback_address: str, callback_function_id: Union[bytes, str], nonce: int, data_version: int, data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (sender, payment, spec_id, callback_address, callback_function_id, nonce, data_version, data) = self.validate_and_normalize_inputs(sender, payment, spec_id, callback_address, callback_function_id, nonce, data_version, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(sender, payment, spec_id, callback_address, callback_function_id, nonce, data_version, data).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, sender: str, payment: int, spec_id: Union[bytes, str], callback_address: str, callback_function_id: Union[bytes, str], nonce: int, data_version: int, data: Union[bytes, str], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (sender, payment, spec_id, callback_address, callback_function_id, nonce, data_version, data) = self.validate_and_normalize_inputs(sender, payment, spec_id, callback_address, callback_function_id, nonce, data_version, data)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(sender, payment, spec_id, callback_address, callback_function_id, nonce, data_version, data).estimateGas(tx_params.as_dict())


class OwnerMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the owner method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("owner")

    def block_call(self, debug: bool = False) -> str:
        _fn = self._underlying_method()
        returned = _fn.call({
            'from': self._operate
        })
        return str(returned)

    def block_send(self, _gaswei: int, _pricewei: int, _valeth: int = 0, _debugtx: bool = False, _receipList: bool = False) -> str:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method()
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': _gaswei,
                'gasPrice': _pricewei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if _debugtx:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if _receipList is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if _debugtx:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET} - broadcast hash")

            if _receipList is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: owner")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, owner: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, owner. Reason: Unknown")

    def send_transaction(self, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())


class SetFulfillmentPermissionMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the setFulfillmentPermission method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("setFulfillmentPermission")

    def validate_and_normalize_inputs(self, node: str, allowed: bool) -> any:
        """Validate the inputs to the setFulfillmentPermission method."""
        self.validator.assert_valid(
            method_name='setFulfillmentPermission',
            parameter_name='_node',
            argument_value=node,
        )
        node = self.validate_and_checksum_address(node)
        self.validator.assert_valid(
            method_name='setFulfillmentPermission',
            parameter_name='_allowed',
            argument_value=allowed,
        )
        return (node, allowed)

    def block_send(self, node: str, allowed: bool, _gaswei: int, _pricewei: int, _valeth: int = 0, _debugtx: bool = False, _receipList: bool = False) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method(node, allowed)
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': _gaswei,
                'gasPrice': _pricewei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if _debugtx:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if _receipList is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if _debugtx:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET} - broadcast hash")

            if _receipList is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: set_fulfillment_permission")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, set_fulfillment_permission: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, set_fulfillment_permission. Reason: Unknown")

    def send_transaction(self, node: str, allowed: bool, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (node, allowed) = self.validate_and_normalize_inputs(node, allowed)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(node, allowed).transact(tx_params.as_dict())

    def build_transaction(self, node: str, allowed: bool, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (node, allowed) = self.validate_and_normalize_inputs(node, allowed)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(node, allowed).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, node: str, allowed: bool, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (node, allowed) = self.validate_and_normalize_inputs(node, allowed)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(node, allowed).estimateGas(tx_params.as_dict())


class TransferOwnershipMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the transferOwnership method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("transferOwnership")

    def validate_and_normalize_inputs(self, new_owner: str) -> any:
        """Validate the inputs to the transferOwnership method."""
        self.validator.assert_valid(
            method_name='transferOwnership',
            parameter_name='newOwner',
            argument_value=new_owner,
        )
        new_owner = self.validate_and_checksum_address(new_owner)
        return (new_owner)

    def block_send(self, new_owner: str, _gaswei: int, _pricewei: int, _valeth: int = 0, _debugtx: bool = False, _receipList: bool = False) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method(new_owner)
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': _gaswei,
                'gasPrice': _pricewei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if _debugtx:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if _receipList is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if _debugtx:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET} - broadcast hash")

            if _receipList is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: transfer_ownership")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, transfer_ownership: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, transfer_ownership. Reason: Unknown")

    def send_transaction(self, new_owner: str, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (new_owner) = self.validate_and_normalize_inputs(new_owner)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(new_owner).transact(tx_params.as_dict())

    def build_transaction(self, new_owner: str, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (new_owner) = self.validate_and_normalize_inputs(new_owner)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(new_owner).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, new_owner: str, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (new_owner) = self.validate_and_normalize_inputs(new_owner)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(new_owner).estimateGas(tx_params.as_dict())


class WithdrawMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the withdraw method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("withdraw")

    def validate_and_normalize_inputs(self, recipient: str, amount: int) -> any:
        """Validate the inputs to the withdraw method."""
        self.validator.assert_valid(
            method_name='withdraw',
            parameter_name='_recipient',
            argument_value=recipient,
        )
        recipient = self.validate_and_checksum_address(recipient)
        self.validator.assert_valid(
            method_name='withdraw',
            parameter_name='_amount',
            argument_value=amount,
        )
        # safeguard against fractional inputs
        amount = int(amount)
        return (recipient, amount)

    def block_send(self, recipient: str, amount: int, _gaswei: int, _pricewei: int, _valeth: int = 0, _debugtx: bool = False, _receipList: bool = False) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method(recipient, amount)
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': _gaswei,
                'gasPrice': _pricewei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if _debugtx:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if _receipList is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if _debugtx:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET} - broadcast hash")

            if _receipList is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: withdraw")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, withdraw: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, withdraw. Reason: Unknown")

    def send_transaction(self, recipient: str, amount: int, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (recipient, amount) = self.validate_and_normalize_inputs(recipient, amount)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(recipient, amount).transact(tx_params.as_dict())

    def build_transaction(self, recipient: str, amount: int, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (recipient, amount) = self.validate_and_normalize_inputs(recipient, amount)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(recipient, amount).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, recipient: str, amount: int, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (recipient, amount) = self.validate_and_normalize_inputs(recipient, amount)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(recipient, amount).estimateGas(tx_params.as_dict())


class WithdrawableMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the withdrawable method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("withdrawable")

    def block_call(self, debug: bool = False) -> int:
        _fn = self._underlying_method()
        returned = _fn.call({
            'from': self._operate
        })
        return int(returned)

    def block_send(self, _gaswei: int, _pricewei: int, _valeth: int = 0, _debugtx: bool = False, _receipList: bool = False) -> int:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method()
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': _gaswei,
                'gasPrice': _pricewei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if _debugtx:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if _receipList is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.waitForTransactionReceipt(txHash)
                    if _debugtx:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                print(f"======== TX blockHash ✅")
                if tx_receipt is not None:
                    print(f"{Bolors.OK}{tx_receipt.blockHash.hex()}{Bolors.RESET}")
                else:
                    print(f"{Bolors.WARNING}{txHash.hex()}{Bolors.RESET} - broadcast hash")

            if _receipList is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: withdrawable")

        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, withdrawable: {message}")
            else:
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, withdrawable. Reason: Unknown")

    def send_transaction(self, tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().transact(tx_params.as_dict())

    def build_transaction(self, tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().buildTransaction(tx_params.as_dict())

    def estimate_gas(self, tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method().estimateGas(tx_params.as_dict())


class SignatureGenerator(Signatures):
    """
        The signature is generated for this and it is installed.
    """

    def __init__(self, abi: any):
        super().__init__(abi)

    def expiry_time(self) -> str:
        return self._function_signatures["EXPIRY_TIME"]

    def cancel_oracle_request(self) -> str:
        return self._function_signatures["cancelOracleRequest"]

    def fulfill_oracle_request(self) -> str:
        return self._function_signatures["fulfillOracleRequest"]

    def get_authorization_status(self) -> str:
        return self._function_signatures["getAuthorizationStatus"]

    def get_chainlink_token(self) -> str:
        return self._function_signatures["getChainlinkToken"]

    def is_owner(self) -> str:
        return self._function_signatures["isOwner"]

    def on_token_transfer(self) -> str:
        return self._function_signatures["onTokenTransfer"]

    def oracle_request(self) -> str:
        return self._function_signatures["oracleRequest"]

    def owner(self) -> str:
        return self._function_signatures["owner"]

    def set_fulfillment_permission(self) -> str:
        return self._function_signatures["setFulfillmentPermission"]

    def transfer_ownership(self) -> str:
        return self._function_signatures["transferOwnership"]

    def withdraw(self) -> str:
        return self._function_signatures["withdraw"]

    def withdrawable(self) -> str:
        return self._function_signatures["withdrawable"]


# pylint: disable=too-many-public-methods,too-many-instance-attributes
class Oracle(ContractBase):
    """Wrapper class for Oracle Solidity contract.

    All method parameters of type `bytes`:code: should be encoded as UTF-8,
    which can be accomplished via `str.encode("utf_8")`:code:.
    """
    _fn_expiry_time: ExpiryTimeMethod
    """Constructor-initialized instance of
    :class:`ExpiryTimeMethod`.
    """

    _fn_cancel_oracle_request: CancelOracleRequestMethod
    """Constructor-initialized instance of
    :class:`CancelOracleRequestMethod`.
    """

    _fn_fulfill_oracle_request: FulfillOracleRequestMethod
    """Constructor-initialized instance of
    :class:`FulfillOracleRequestMethod`.
    """

    _fn_get_authorization_status: GetAuthorizationStatusMethod
    """Constructor-initialized instance of
    :class:`GetAuthorizationStatusMethod`.
    """

    _fn_get_chainlink_token: GetChainlinkTokenMethod
    """Constructor-initialized instance of
    :class:`GetChainlinkTokenMethod`.
    """

    _fn_is_owner: IsOwnerMethod
    """Constructor-initialized instance of
    :class:`IsOwnerMethod`.
    """

    _fn_on_token_transfer: OnTokenTransferMethod
    """Constructor-initialized instance of
    :class:`OnTokenTransferMethod`.
    """

    _fn_oracle_request: OracleRequestMethod
    """Constructor-initialized instance of
    :class:`OracleRequestMethod`.
    """

    _fn_owner: OwnerMethod
    """Constructor-initialized instance of
    :class:`OwnerMethod`.
    """

    _fn_set_fulfillment_permission: SetFulfillmentPermissionMethod
    """Constructor-initialized instance of
    :class:`SetFulfillmentPermissionMethod`.
    """

    _fn_transfer_ownership: TransferOwnershipMethod
    """Constructor-initialized instance of
    :class:`TransferOwnershipMethod`.
    """

    _fn_withdraw: WithdrawMethod
    """Constructor-initialized instance of
    :class:`WithdrawMethod`.
    """

    _fn_withdrawable: WithdrawableMethod
    """Constructor-initialized instance of
    :class:`WithdrawableMethod`.
    """

    SIGNATURES: SignatureGenerator = None

    def __init__(
            self,
            core_lib: MiliDoS,
            contract_address: str,
            validator: OracleValidator = None,
    ):
        """Get an instance of wrapper for smart contract.
        """
        # pylint: disable=too-many-statements
        super().__init__()
        self.contract_address = contract_address
        web3 = core_lib.w3

        if not validator:
            validator = OracleValidator(web3, contract_address)

        # if any middleware was imported, inject it
        try:
            MIDDLEWARE
        except NameError:
            pass
        else:
            try:
                for middleware in MIDDLEWARE:
                    web3.middleware_onion.inject(
                        middleware['function'], layer=middleware['layer'],
                    )
            except ValueError as value_error:
                if value_error.args == ("You can't add the same un-named instance twice",):
                    pass

        self._web3_eth = web3.eth
        functions = self._web3_eth.contract(address=to_checksum_address(contract_address), abi=Oracle.abi()).functions
        signed = SignatureGenerator(Oracle.abi())
        validator.bindSignatures(signed)
        self.SIGNATURES = signed
        self._fn_expiry_time = ExpiryTimeMethod(core_lib, contract_address, functions.EXPIRY_TIME, validator)
        self._fn_cancel_oracle_request = CancelOracleRequestMethod(core_lib, contract_address, functions.cancelOracleRequest, validator)
        self._fn_fulfill_oracle_request = FulfillOracleRequestMethod(core_lib, contract_address, functions.fulfillOracleRequest, validator)
        self._fn_get_authorization_status = GetAuthorizationStatusMethod(core_lib, contract_address, functions.getAuthorizationStatus, validator)
        self._fn_get_chainlink_token = GetChainlinkTokenMethod(core_lib, contract_address, functions.getChainlinkToken, validator)
        self._fn_is_owner = IsOwnerMethod(core_lib, contract_address, functions.isOwner, validator)
        self._fn_on_token_transfer = OnTokenTransferMethod(core_lib, contract_address, functions.onTokenTransfer, validator)
        self._fn_oracle_request = OracleRequestMethod(core_lib, contract_address, functions.oracleRequest, validator)
        self._fn_owner = OwnerMethod(core_lib, contract_address, functions.owner, validator)
        self._fn_set_fulfillment_permission = SetFulfillmentPermissionMethod(core_lib, contract_address, functions.setFulfillmentPermission, validator)
        self._fn_transfer_ownership = TransferOwnershipMethod(core_lib, contract_address, functions.transferOwnership, validator)
        self._fn_withdraw = WithdrawMethod(core_lib, contract_address, functions.withdraw, validator)
        self._fn_withdrawable = WithdrawableMethod(core_lib, contract_address, functions.withdrawable, validator)

    def event_cancel_oracle_request(
            self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """
        Implementation of event cancel_oracle_request in contract Oracle
        Get log entry for CancelOracleRequest event.
                :param tx_hash: hash of transaction emitting CancelOracleRequest event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=Oracle.abi()).events.CancelOracleRequest().processReceipt(tx_receipt)

    def event_oracle_request(
            self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """
        Implementation of event oracle_request in contract Oracle
        Get log entry for OracleRequest event.
                :param tx_hash: hash of transaction emitting OracleRequest event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=Oracle.abi()).events.OracleRequest().processReceipt(tx_receipt)

    def event_ownership_transferred(
            self, tx_hash: Union[HexBytes, bytes]
    ) -> Tuple[AttributeDict]:
        """
        Implementation of event ownership_transferred in contract Oracle
        Get log entry for OwnershipTransferred event.
                :param tx_hash: hash of transaction emitting OwnershipTransferred event
        """
        tx_receipt = self._web3_eth.getTransactionReceipt(tx_hash)
        return self._web3_eth.contract(address=to_checksum_address(self.contract_address), abi=Oracle.abi()).events.OwnershipTransferred().processReceipt(tx_receipt)

    def expiry_time(self) -> int:
        """
        Implementation of expiry_time in contract Oracle
        Method of the function
    
    
    
        """

        return self._fn_expiry_time.block_call()

    def cancel_oracle_request(self, request_id: Union[bytes, str], payment: int, callback_func: Union[bytes, str], expiration: int) -> None:
        """
        Implementation of cancel_oracle_request in contract Oracle
        Method of the function
    
    
    
        """

        return self._fn_cancel_oracle_request.block_send(request_id, payment, callback_func, expiration, self.call_contract_fee_amount, self.call_contract_fee_price, 0, self.call_contract_debug_flag, self.call_contract_enforce_tx_receipt)

    def fulfill_oracle_request(self, request_id: Union[bytes, str], payment: int, callback_address: str, callback_function_id: Union[bytes, str], expiration: int, data: Union[bytes, str]) -> bool:
        """
        Implementation of fulfill_oracle_request in contract Oracle
        Method of the function
    
    
    
        """

        return self._fn_fulfill_oracle_request.block_send(request_id, payment, callback_address, callback_function_id, expiration, data, self.call_contract_fee_amount, self.call_contract_fee_price, 0, self.call_contract_debug_flag, self.call_contract_enforce_tx_receipt)

    def get_authorization_status(self, node: str) -> bool:
        """
        Implementation of get_authorization_status in contract Oracle
        Method of the function
    
    
    
        """

        return self._fn_get_authorization_status.block_call(node)

    def get_chainlink_token(self) -> str:
        """
        Implementation of get_chainlink_token in contract Oracle
        Method of the function
    
    
    
        """

        return self._fn_get_chainlink_token.block_call()

    def is_owner(self) -> bool:
        """
        Implementation of is_owner in contract Oracle
        Method of the function
    
    
    
        """

        return self._fn_is_owner.block_call()

    def on_token_transfer(self, sender: str, amount: int, data: Union[bytes, str]) -> None:
        """
        Implementation of on_token_transfer in contract Oracle
        Method of the function
    
    
    
        """

        return self._fn_on_token_transfer.block_send(sender, amount, data, self.call_contract_fee_amount, self.call_contract_fee_price, 0, self.call_contract_debug_flag, self.call_contract_enforce_tx_receipt)

    def oracle_request(self, sender: str, payment: int, spec_id: Union[bytes, str], callback_address: str, callback_function_id: Union[bytes, str], nonce: int, data_version: int, data: Union[bytes, str]) -> None:
        """
        Implementation of oracle_request in contract Oracle
        Method of the function
    
    
    
        """

        return self._fn_oracle_request.block_send(sender, payment, spec_id, callback_address, callback_function_id, nonce, data_version, data, self.call_contract_fee_amount, self.call_contract_fee_price, 0, self.call_contract_debug_flag, self.call_contract_enforce_tx_receipt)

    def owner(self) -> str:
        """
        Implementation of owner in contract Oracle
        Method of the function
    
    
    
        """

        return self._fn_owner.block_call()

    def set_fulfillment_permission(self, node: str, allowed: bool) -> None:
        """
        Implementation of set_fulfillment_permission in contract Oracle
        Method of the function
    
    
    
        """

        return self._fn_set_fulfillment_permission.block_send(node, allowed, self.call_contract_fee_amount, self.call_contract_fee_price, 0, self.call_contract_debug_flag, self.call_contract_enforce_tx_receipt)

    def transfer_ownership(self, new_owner: str) -> None:
        """
        Implementation of transfer_ownership in contract Oracle
        Method of the function
    
    
    
        """

        return self._fn_transfer_ownership.block_send(new_owner, self.call_contract_fee_amount, self.call_contract_fee_price, 0, self.call_contract_debug_flag, self.call_contract_enforce_tx_receipt)

    def withdraw(self, recipient: str, amount: int) -> None:
        """
        Implementation of withdraw in contract Oracle
        Method of the function
    
    
    
        """

        return self._fn_withdraw.block_send(recipient, amount, self.call_contract_fee_amount, self.call_contract_fee_price, 0, self.call_contract_debug_flag, self.call_contract_enforce_tx_receipt)

    def withdrawable(self) -> int:
        """
        Implementation of withdrawable in contract Oracle
        Method of the function
    
    
    
        """

        return self._fn_withdrawable.block_call()

    def CallContractWait(self, t_long: int) -> "Oracle":
        self._fn_expiry_time.setWait(t_long)
        self._fn_cancel_oracle_request.setWait(t_long)
        self._fn_fulfill_oracle_request.setWait(t_long)
        self._fn_get_authorization_status.setWait(t_long)
        self._fn_get_chainlink_token.setWait(t_long)
        self._fn_is_owner.setWait(t_long)
        self._fn_on_token_transfer.setWait(t_long)
        self._fn_oracle_request.setWait(t_long)
        self._fn_owner.setWait(t_long)
        self._fn_set_fulfillment_permission.setWait(t_long)
        self._fn_transfer_ownership.setWait(t_long)
        self._fn_withdraw.setWait(t_long)
        self._fn_withdrawable.setWait(t_long)
        return self

    @staticmethod
    def abi():
        """Return the ABI to the underlying contract."""
        return json.loads(
            '[{"inputs":[{"internalType":"address","name":"_link","type":"address"}],"stateMutability":"nonpayable","type":"constructor"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"requestId","type":"bytes32"}],"name":"CancelOracleRequest","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"bytes32","name":"specId","type":"bytes32"},{"indexed":false,"internalType":"address","name":"requester","type":"address"},{"indexed":false,"internalType":"bytes32","name":"requestId","type":"bytes32"},{"indexed":false,"internalType":"uint256","name":"payment","type":"uint256"},{"indexed":false,"internalType":"address","name":"callbackAddr","type":"address"},{"indexed":false,"internalType":"bytes4","name":"callbackFunctionId","type":"bytes4"},{"indexed":false,"internalType":"uint256","name":"cancelExpiration","type":"uint256"},{"indexed":false,"internalType":"uint256","name":"dataVersion","type":"uint256"},{"indexed":false,"internalType":"bytes","name":"data","type":"bytes"}],"name":"OracleRequest","type":"event"},{"anonymous":false,"inputs":[{"indexed":true,"internalType":"address","name":"previousOwner","type":"address"},{"indexed":true,"internalType":"address","name":"newOwner","type":"address"}],"name":"OwnershipTransferred","type":"event"},{"inputs":[],"name":"EXPIRY_TIME","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes32","name":"_requestId","type":"bytes32"},{"internalType":"uint256","name":"_payment","type":"uint256"},{"internalType":"bytes4","name":"_callbackFunc","type":"bytes4"},{"internalType":"uint256","name":"_expiration","type":"uint256"}],"name":"cancelOracleRequest","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bytes32","name":"_requestId","type":"bytes32"},{"internalType":"uint256","name":"_payment","type":"uint256"},{"internalType":"address","name":"_callbackAddress","type":"address"},{"internalType":"bytes4","name":"_callbackFunctionId","type":"bytes4"},{"internalType":"uint256","name":"_expiration","type":"uint256"},{"internalType":"bytes32","name":"_data","type":"bytes32"}],"name":"fulfillOracleRequest","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"_node","type":"address"}],"name":"getAuthorizationStatus","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"getChainlinkToken","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"isOwner","outputs":[{"internalType":"bool","name":"","type":"bool"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"address","name":"_sender","type":"address"},{"internalType":"uint256","name":"_amount","type":"uint256"},{"internalType":"bytes","name":"_data","type":"bytes"}],"name":"onTokenTransfer","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"_sender","type":"address"},{"internalType":"uint256","name":"_payment","type":"uint256"},{"internalType":"bytes32","name":"_specId","type":"bytes32"},{"internalType":"address","name":"_callbackAddress","type":"address"},{"internalType":"bytes4","name":"_callbackFunctionId","type":"bytes4"},{"internalType":"uint256","name":"_nonce","type":"uint256"},{"internalType":"uint256","name":"_dataVersion","type":"uint256"},{"internalType":"bytes","name":"_data","type":"bytes"}],"name":"oracleRequest","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[],"name":"owner","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"address","name":"_node","type":"address"},{"internalType":"bool","name":"_allowed","type":"bool"}],"name":"setFulfillmentPermission","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"newOwner","type":"address"}],"name":"transferOwnership","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"_recipient","type":"address"},{"internalType":"uint256","name":"_amount","type":"uint256"}],"name":"withdraw","outputs":[],"stateMutability":"nonpayable","type":"function"},{"inputs":[],"name":"withdrawable","outputs":[{"internalType":"uint256","name":"","type":"uint256"}],"stateMutability":"view","type":"function"}]'
            # noqa: E501 (line-too-long)
        )

# pylint: disable=too-many-lines



from . import Config


def BSCTest() -> Config:
    return Config(
        network_name="BSCTestnet",
        rpc_url="https://data-seed-prebsc-1-s1.binance.org:8545",
        chain_id=97,
        symbol="BNB",
        block_explorer="https://testnet.bscscan.com/",
        link_token="0x84b9B910527Ad5C03A9Ca831909E21e236EA7b06",
        link_vrf_coordinator="0xa555fC018435bef5A13C6c6870a9d4C11DEC329C",
        link_keyhash="0xcaf3c3727e033261d383b315559476f48034c13b18f8cafed4d871abe5049186"
    )


def Ethereum() -> Config:
    return Config(
        network_name="EthereumMainnet",
        rpc_url="https://mainnet.infura.io/v3/9aa3d95b3bc440fa88ea12eaa4456161",
        chain_id=1,
        symbol="ETH",
        block_explorer="https://etherscan.io/",
        link_token="0x514910771AF9Ca656af840dff83E8264EcF986CA",
        link_vrf_coordinator="0xf0d54349aDdcf704F77AE15b96510dEA15cb7952",
        link_keyhash="0xAA77729D3466CA35AE8D28B3BBAC7CC36A5031EFDC430821C02BC31A238AF445"
    )


def EthereumFL() -> Config:
    return Config(
        network_name="EthereumMainnet",
        rpc_url="https://rpc.flashbots.net",
        chain_id=1,
        symbol="ETH",
        block_explorer="https://flashbots-explorer.marto.lol/",
        link_token="0x514910771AF9Ca656af840dff83E8264EcF986CA",
        link_vrf_coordinator="0xf0d54349aDdcf704F77AE15b96510dEA15cb7952",
        link_keyhash="0xAA77729D3466CA35AE8D28B3BBAC7CC36A5031EFDC430821C02BC31A238AF445"
    )


def GoerliTestnetFL() -> Config:
    return Config(
        network_name="GoerliTestnet",
        rpc_url="https://rpc-goerli.flashbots.net",
        chain_id=5,
        symbol="ETH",
        block_explorer="https://flashbots-explorer.marto.lol/",
        link_token="0x514910771AF9Ca656af840dff83E8264EcF986CA",
        link_vrf_coordinator="0xf0d54349aDdcf704F77AE15b96510dEA15cb7952",
        link_keyhash="0xAA77729D3466CA35AE8D28B3BBAC7CC36A5031EFDC430821C02BC31A238AF445"
    )


def GoerliTestnet() -> Config:
    return Config(
        network_name="GoerliTestnet",
        rpc_url="https://goerli.infura.io/v3/9aa3d95b3bc440fa88ea12eaa4456161",
        chain_id=5,
        symbol="ETH",
        block_explorer="https://goerli.etherscan.io"
    )


def BSCMain() -> Config:
    return Config(
        network_name="BinanceMainnet",
        rpc_url="https://bsc-dataseed1.binance.org",
        chain_id=56,
        symbol="BNB",
        block_explorer="https://bscscan.com",
        link_token="0x404460C6A5EdE2D891e8297795264fDe62ADBB75",
        link_keyhash="0xc251acd21ec4fb7f31bb8868288bfdbaeb4fbfec2df3735ddbd4f7dc8d60103c",
        link_vrf_coordinator="0x747973a5A2a4Ae1D3a8fDF5479f1514F65Db9C31"
    )


def PolygonMumbai() -> Config:
    return Config(
        network_name="PolygonMumbaiTestnet",
        rpc_url="https://rpc-mumbai.maticvigil.com",
        chain_id=80001,
        symbol="MATIC",
        block_explorer="https://mumbai.polygonscan.com/",
        link_token="0x326C977E6efc84E512bB9C30f76E30c160eD06FB",
        link_keyhash="0x6e75b569a01ef56d18cab6a8e71e6600d6ce853834d4a5748b720d06f878b3a4",
        link_vrf_coordinator="0x8C7382F9D8f56b33781fE506E897a4F1e2d17255"
    )


def PolygonMainnet() -> Config:
    return Config(
        network_name="PolygonMainnet",
        rpc_url="https://polygon-rpc.com",
        chain_id=137,
        symbol="MATIC",
        block_explorer="https://explorer.matic.network/",
        link_token="0xb0897686c545045aFc77CF20eC7A532E3120E0F1",
        link_keyhash="0xf86195cf7690c55907b2b611ebb7343a6f649bff128701cc542f0569e2c549da",
        link_vrf_coordinator="0x3d2341ADb2D31f1c5530cDC622016af293177AE0"
    )


def KovanTestnet() -> Config:
    return Config(
        network_name="KovanTestnet",
        rpc_url="https://kovan.infura.io/v3/2019a99711c648f8951a640e8031ca33",
        chain_id=42,
        symbol="KETH",
        block_explorer="https://kovan.etherscan.io/",
        link_token="0xa36085F69e2889c224210F603D836748e7dC0088",
        link_vrf_coordinator="0xdD3782915140c8f3b190B5D67eAc6dc5760C46E9",
        link_keyhash="0x6c3699283bda56ad74f6b855546325b68d482e983852a7a82979cc4807b641f4"
    )


def RopstenTestnet() -> Config:
    return Config(
        network_name="RopstenTestnet",
        rpc_url="https://ropsten.infura.io/v3/9aa3d95b3bc440fa88ea12eaa4456161",
        chain_id=3,
        symbol="RETH",
        block_explorer="https://ropsten.etherscan.io/"
    )


def RinkebyTestnet() -> Config:
    return Config(
        network_name="RinkebyTestnet",
        rpc_url="https://rinkeby.infura.io/v3/9aa3d95b3bc440fa88ea12eaa4456161",
        chain_id=4,
        symbol="ETH",
        block_explorer="https://rinkeby.etherscan.io/"
    )


def MoonBeamTestnet() -> Config:
    return Config(
        # wss://wss.testnet.moonbeam.network
        network_name="MoonbaseAlphanet",
        rpc_url="https://rpc.testnet.moonbeam.network",
        chain_id=1287,
        symbol="DEV",
        block_explorer="https://moonbase-blockscout.testnet.moonbeam.network/"
    )


def HuobiChainMainnet() -> Config:
    return Config(
        network_name="Heco",
        rpc_url="https://http-mainnet.hecochain.com",
        chain_id=128,
        symbol="HT",
        block_explorer="https://hecoinfo.com/"
    )


def HuobiChainTestnet() -> Config:
    return Config(
        network_name="HecoTest",
        rpc_url="https://http-testnet.hecochain.com",
        chain_id=256,
        symbol="HT",
        block_explorer="https://testnet.hecoinfo.com/"
    )


def XDaiMainnet() -> Config:
    """
    more resources on
    https://www.xdaichain.com/for-developers/developer-resources#json-rpc-endpoints
    :return:
    """
    return Config(
        network_name="xDaiStake",
        rpc_url="https://rpc.xdaichain.com",
        # rpc_url="https://xdai.poanetwork.dev",
        # rpc_url="https://stake.getblock.io/mainnet/?api_key=bc690eca-e18a-4c53-b8e9-0f413e225e69",
        chain_id=100,
        symbol="xDai",
        block_explorer="https://explorer.anyblock.tools/ethereum/poa/xdai/",
        bridge="https://bridge.xdaichain.com/"
    )


def OKChainMainnet() -> Config:
    return Config(
        network_name="OKExChainMainnet",
        rpc_url="https://exchainrpc.okex.org",
        chain_id=66,
        symbol="OKT",
        block_explorer="https://www.oklink.com/okexchain/"
    )


def OKChainTestnet() -> Config:
    return Config(
        network_name="OKExChainTestnet",
        rpc_url="https://exchaintestrpc.okex.org",
        chain_id=65,
        symbol="OKT",
        block_explorer="https://www.oklink.com/okexchain-test/"
    )


def RSCMainnet() -> Config:
    return Config(
        network_name="RSCMainnet",
        rpc_url="https://rpc-mainnet.raisc.io",
        chain_id=1023,
        symbol="RSC",
        block_explorer="http://explorer.raisc.io/",
        link_token="0x79e93714E4049Ae8745F1FA21838d6FdfDCa9cA6",
        link_vrf_coordinator="0x3c89bdbfF0B1a27c126b89e78F21a7B66DA2dbeC",
        link_keyhash="0x66c8089c4313b58a4344f69f796705ea1a7b2c180836e56f4785bde84f7af32f"
    )


def OasisMainnet() -> Config:
    return Config(
        network_name="OasisChain",
        rpc_url="https://rpc1.oasischain.io",
        chain_id=1023,
        symbol="OAC",
        block_explorer="https://scan.oasischain.io/"
    )


def AvalancheFujiTestnet() -> Config:
    return Config(
        network_name="AvalancheTestnet",
        rpc_url="https://api.avax-test.network/ext/bc/C/rpc",
        chain_id=43113,
        symbol="AVAX",
        block_explorer="https://testnet.snowtrace.io/"
    )


def AvalancheMainnet() -> Config:
    return Config(
        network_name="AvalancheMainnet",
        rpc_url="https://api.avax.network/ext/bc/C/rpc",
        chain_id=43114,
        symbol="AVAX",
        block_explorer="https://snowtrace.io/"
    )


def OptimisticEthereum() -> Config:
    return Config(
        network_name="OptimisticEthereumMainnet",
        rpc_url="https://mainnet.optimism.io",
        chain_id=10,
        symbol="ETH",
        block_explorer="https://optimistic.etherscan.io/"
    )


def OptimisticEthereumTestnet() -> Config:
    return Config(
        network_name="OptimisticEthereumMainnet",
        rpc_url="https://kovan.optimism.io",
        chain_id=69,
        symbol="ETH",
        block_explorer="https://kovan-optimistic.etherscan.io/"
    )


# https://github.com/aurora-is-near/aurora-engine
def NearAuroraTestnet() -> Config:
    return Config(
        network_name="AuroraTestnet",
        rpc_url="https://testnet.aurora.dev",
        chain_id=1313161555,
        symbol="NEAR",
        block_explorer="https://explorer.testnet.near.org/"
    )


def NearAuroraMainnet() -> Config:
    return Config(
        network_name="AuroraTestnet",
        rpc_url="https://rpc.ankr.com/near",
        chain_id=1313161554,
        symbol="NEAR",
        block_explorer="https://explorer.near.org/"
    )


def ArbitrumOne() -> Config:
    return Config(
        network_name="ArbitrumOne",
        rpc_url="https://rpc.ankr.com/arbitrum",
        chain_id=42161,
        symbol="ETH",
        block_explorer="https://arbiscan.io/"
    )


def FantomOperaMainnetRPC() -> Config:
    return Config(
        network_name="FantomOpera",
        rpc_url="https://rpc.ftm.tools",
        chain_id=250,
        symbol="FTM",
        block_explorer="https://ftmscan.com/"
    )


def FantomTestnetRPC() -> Config:
    return Config(
        network_name="FantomTestnet",
        rpc_url="https://rpc.testnet.fantom.network",
        chain_id=4002,
        symbol="FTM",
        block_explorer="https://testnet.ftmscan.com/"
    )


def HarmonyMainnet() -> Config:
    return Config(
        network_name="HarmonyMainnet",
        rpc_url="https://api.harmony.one",
        chain_id=1666600000,
        symbol="ONE",
        block_explorer="https://explorer.harmony.one/"
    )


def HarmonyTestnet() -> Config:
    return Config(
        network_name="HarmonyTestnet",
        rpc_url="https://api.s0.b.hmny.io",
        chain_id=1666700000,
        symbol="ONE",
        block_explorer="https://explorer.harmony.one/"
    )


def Kucoin() -> Config:
    return Config(
        network_name="KucoinMainnet",
        rpc_url="https://rpc-mainnet.kcc.network",
        chain_id=321,
        symbol="KCS",
        block_explorer="https://scan.kcc.network/"
    )


def KucoinTestnet() -> Config:
    return Config(
        network_name="KucoinTestnet",
        rpc_url="https://scan-testnet.kcc.network",
        chain_id=322,
        symbol="KCS",
        block_explorer="https://scan-testnet.kcc.network/"
    )


def Elastos() -> Config:
    return Config(
        network_name="Elastos",
        rpc_url="https://api.elastos.io/eth",
        chain_id=20,
        symbol="KCS",
        block_explorer="https://explorer.elaeth.io/"
    )


def Gnosis() -> Config:
    return Config(
        network_name="Gnosis",
        rpc_url="https://rpc.gnosischain.com",
        chain_id=100,
        symbol="XDAI",
        block_explorer="https://blockscout.com/xdai/mainnet/"
    )


def Ubiq() -> Config:
    return Config(
        network_name="Ubiq",
        rpc_url="https://rpc.octano.dev",
        chain_id=8,
        symbol="UBQ",
        block_explorer="https://ubiqscan.io/"
    )


def Cronos() -> Config:
    return Config(
        network_name="Cronos",
        rpc_url="https://evm-cronos.crypto.org",
        chain_id=25,
        symbol="CRO",
        block_explorer="https://cronos.crypto.org/explorer/"
    )


def Solana() -> Config:
    return Config(
        network_name="SOL",
        rpc_url="https://rpc.ankr.com/solana",
        chain_id=25221,
        symbol="CRO",
        block_explorer="https://solscan.io/"
    )


def SmartBCH() -> Config:
    return Config(
        network_name="BCHMainnet",
        rpc_url="https://smartbch.fountainhead.cash/mainnet",
        chain_id=10000,
        symbol="BCH",
        block_explorer="https://www.smartscan.ca.sh/"
    )


def CeloMainnet() -> Config:
    return Config(
        network_name="CeloMainnet",
        rpc_url="https://rpc.ankr.com/celo",
        chain_id=42220,
        symbol="CELO",
        block_explorer="https://celoscan.com/"
    )



from typing import List, Any, Callable, Dict, Optional, Union
from hexbytes import HexBytes

import hashlib


class MerkleTree:
    """
    Merkle Tree Lx.

    """

    def __init__(
            self, leaves: List[Any], hash_function: Callable, sort: Optional[bool] = False
    ) -> None:
        self.hash_function = self.bufferify_function(hash_function)
        self.leaves = leaves
        self.sort_leaves = sort
        self.sort_pairs = sort
        self._process_leaves()

    @staticmethod
    def to_hex(value: Any) -> str:
        return value.hex()

    @staticmethod
    def bufferify(value: str) -> str:
        if type(value) == bytes:
            return value
        else:
            return value.encode()

    @staticmethod
    def bufferify_function(func: Callable) -> Callable:
        def f(val):
            return MerkleTree.bufferify(func(val))

        return f

    def _process_leaves(self) -> None:
        self.leaves = [self.bufferify(leaf) for leaf in self.leaves]
        if self.sort_leaves:
            self.leaves.sort()
        self.layers = [self.leaves]
        self._create_hashes(self.leaves)

    def _create_hashes(self, nodes: List[Any]) -> None:
        while len(nodes) > 1:
            n = len(nodes)
            layer_index = len(self.layers)
            self.layers.append([])
            for i in range(0, n, 2):
                if n == i + 1 and n % 2 == 1:
                    self.layers[layer_index].append(nodes[i])
                    continue
                left = self.bufferify(nodes[i])
                right = left if i + 1 == n else self.bufferify(nodes[i + 1])
                combined = (
                    left + right
                    if not (self.sort_pairs and right > left)
                    else right + left
                )
                hashed_data = self.hash_function(combined)
                self.layers[layer_index].append(hashed_data)
            nodes = self.layers[layer_index]

    def get_hex_layers(self) -> List[str]:
        return [[self.to_hex(leaf) for leaf in layer] for layer in self.layers]

    def get_root(self) -> any:
        try:
            return self.layers[-1][0]
        except IndexError:
            return []

    def get_hex_root(self) -> any:
        if self.get_root():
            return self.to_hex(self.get_root())
        else:
            return []

    def get_proof(self, leaf: Union[str, bytes], index: Optional[int] = None) -> List[Dict[str, Any]]:
        proof = []
        leaf = self.bufferify(leaf)
        if not index:
            try:
                index = self.leaves.index(leaf)
            except ValueError:
                return []

        for layer in self.layers:
            is_right_node = index % 2
            pair_index = index - 1 if is_right_node else index + 1
            if pair_index < len(layer):
                proof.append(
                    {
                        "position": "left" if is_right_node else "right",
                        "data": layer[pair_index],
                    }
                )
            index = index // 2
        return proof

    def get_hex_proof(self, leaf: Union[str, bytes], index: Optional[int] = 0) -> List[str]:
        return [self.to_hex(item["data"]) for item in self.get_proof(leaf, index)]

    def verify(self, proof: List[Dict[str, str]], target_node: str, root: str) -> bool:
        hash = self.bufferify(target_node)
        root = self.bufferify(root)
        for node in proof:
            data = self.bufferify(node["data"])
            is_left_node = node["position"] == "left"
            if self.sort_pairs:
                combined = data + hash if data <= hash else hash + data
            else:
                combined = data + hash if is_left_node else hash + data
            hash = self.hash_function(combined)
        return hash == root

    def get_depth(self) -> int:
        return len(self.layers) - 1

    def reset_tree(self) -> None:
        self.leaves = []
        self.layers = []


def sha256(x) -> bytes:
    return hashlib.sha256(x).digest()


"""
leaves = [sha256(leaf.encode()) for leaf in "abc"]
tree = MerkleTree(leaves, sha256)
root = tree.get_root()
hex_root = tree.get_hex_root()
leaf = sha256("a".encode())
bad_leaf = sha256("x".encode())
proof = tree.get_proof(leaf)
ok = tree.verify(proof, leaf, root)  # returns True
ok2 = tree.verify(proof, bad_leaf, root)  # returns False
print(hex_root, tree.get_hex_proof(leaf), ok, ok2)

"""


class Kyc:
    def __init__(self):
        self.data = list()
        self._tree = None
        self.root_hex = None
        self.root_layers = None

    def updateList(self, address_list: List[str]) -> "Kyc":
        self.data = address_list
        self._tree = MerkleTree([sha256(leaf.encode()) for leaf in address_list], sha256)
        self.root_hex = self._tree.get_hex_root()
        self.root_layers = self._tree.get_hex_layers()
        return self

    def getKycDataFromAddress(self, address: str) -> list:
        if address not in self.data:
            print(f"Address {address} is not inside the collection")
            return list()
        if self._tree is None:
            print("tree object is not initialized, please updateList the collection first.")
            return list()

        leaf = sha256(address.encode())
        kyc_proof = self._tree.get_hex_proof(leaf)
        proof = self._tree.get_proof(leaf)
        proof_ok = self._tree.verify(proof, leaf, self._tree.get_root())
        print(f"KYC proof ok? {proof_ok}")

        return kyc_proof



from __future__ import print_function

import argparse as ap
import ast
import subprocess
import argparse
import os
import re
import sys

command = "solflatliner"


# solflatliner --o upgrade.sol upgradable/TransparentUpgradeableProxy.sol 0.8.12
class SolflatlinerWrapper:
    """
    This is the solidity flatliner wrapper processor
    """

    def __init__(self, root_space: str):
        self.output_file = False
        self.input_file = False
        self.proccess_result = False
        self.workspace = root_space

    def check_program(self, cmd) -> bool:
        rc = subprocess.call(['which', cmd])
        if rc == 0:
            print(f"{cmd} is installed")
            return True
        else:
            print(f"{cmd} is missing in the system")
            print(f"Try to install this via pip3 install solflatliner")
            return False

    def make_flaten_tmp_file(self, file_path_from_vault: str, version: str):
        if self.proccess_result is True:
            print(f"The subject is already success - {file_path_from_vault}")
            return

            # Split the path in head and tail pair
        head_tail = os.path.split(file_path_from_vault)
        file_name_seg = head_tail[1].split(".")

        if file_name_seg[1] != "sol":
            print("There is no sol file found in the contain")
            return

        if self.check_program(command) is False:
            return

        self.input_file = file_path_from_vault
        space_location = os.path.join(self.workspace, f"{file_name_seg[0]}_flat.sol")
        self.output_file = os.path.join("out", space_location)
        solc_args = [command]
        solc_args += ["--o"]
        solc_args += [space_location]
        solc_args += [file_path_from_vault]
        solc_args += [version]
        print(f"preprocess cmd: {solc_args}")
        solc_proc = subprocess.run(solc_args, stdout=subprocess.PIPE, universal_newlines=True)
        solc_proc.check_returncode()
        self.proccess_result = True



from typing import Tuple, List, Union

from ..bases import ContractBase
from ..multicall2 import Multicall2
from ..multicall.execute import MultiCallerV1Contract
from ...libeb import MiliDoS


class MultiCallerContract(MultiCallerV1Contract):
    """
    Need some tests, help is needed
    """
    def __init__(self, moon: MiliDoS, address: str):
        super().__init__(moon, address)

    def _callersetup(self):
        self.caller: Multicall2 = Multicall2(self._moon, self._address).CallAutoConf(self._moon).CallDebug(True)

    def Config(self) -> Multicall2:
        return self.caller

    def addCallClassic(self, contract: str, abi: str, method: str, params: list) -> "MultiCallerContract":
        super().addCallClassic(contract, abi, method, params)
        return self

    def addCallEasy(self, contract: ContractBase, method: str, params: list) -> "MultiCallerContract":
        super().addCallEasy(contract, method, params)
        return self

    def executeSimple(self) -> Tuple[int, List[Union[bytes, str]]]:
        """
        gives the returns of the block number
        :return:  (blocknumber, hash, results)
        """
        (blocknum, data) = self.caller.aggregate(self._data)
        return (blocknum, data)

    def executeTry(self) -> List[Union[bytes, str]]:
        return self.caller.try_aggregate(True, self._data)

    def executeTryWithBlock(self) -> Tuple[int, Union[bytes, str], List[Union[bytes, str]]]:
        """
        gives the block dat and the results
        :return:  (blocknumber, hash, results)
        """
        return self.caller.try_block_and_aggregate(True, self._data)
